from typing import Optional, Tuple

import pytest  # type:ignore[import]

from ahbicht.edifact import EdifactFormat, pruefidentifikator_to_format


class TestEdifact:
    """
    Tests the edifact module
    """

    @pytest.mark.parametrize(
        "expectation_tuple",
        [
            ("11042", EdifactFormat.UTILMD),
            ("13002", EdifactFormat.MSCONS),
            ("25001", EdifactFormat.UTILTS),
            ("10000", None),
        ],
    )
    def test_pruefi_to_format(self, expectation_tuple: Tuple[str, EdifactFormat]):
        """
        Tests that the prüfis can be mapped to an EDIFACT format
        """
        assert pruefidentifikator_to_format(expectation_tuple[0]) == expectation_tuple[1]

    @pytest.mark.parametrize(
        "illegal_pruefi",
        [None, "", "asdas", "01234"],
    )
    def test_illegal_pruefis(self, illegal_pruefi: Optional[str]):
        """
        Test that illegal pruefis are not accepted
        :return:
        """
        with pytest.raises(ValueError):
            pruefidentifikator_to_format(illegal_pruefi)  # type:ignore[arg-type] # ok, because this raises an error



"""
This module contains a class to store _all_ kinds of content evaluation results.
"""
from typing import Dict, Optional
from uuid import UUID

import attr
from marshmallow import Schema, fields, post_load

from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    EvaluatedFormatConstraint,
    EvaluatedFormatConstraintSchema,
)


# pylint: disable=too-few-public-methods, no-self-use, unused-argument
@attr.s(auto_attribs=True)
class ContentEvaluationResult:
    """
    A class that holds the results of a full content evaluation (meaning all hints, requirement constraints and
    format constraints have been evaluated)
    """

    hints: Dict[str, Optional[str]]  #: maps the key of a hint (e.g. "501" to a hint text)
    format_constraints: Dict[
        str, EvaluatedFormatConstraint
    ]  #: maps the key of a format constraint to the respective evaluation result
    requirement_constraints: Dict[
        str, ConditionFulfilledValue
    ]  #: maps the key of a requirement_constraint to the respective evaluation result
    id: Optional[UUID] = None  #: optional guid


class ContentEvaluationResultSchema(Schema):
    """
    A schema to (de)serialize ContentEvaluationResults
    """

    hints = fields.Dict(keys=fields.String(allow_none=False), values=fields.String(allow_none=True), required=True)
    format_constraints = fields.Dict(
        keys=fields.String(allow_none=False),
        values=fields.Nested(EvaluatedFormatConstraintSchema, allow_none=False, required=True),
        required=True,
    )
    requirement_constraints = fields.Dict(
        keys=fields.String(allow_none=False), values=fields.String(allow_none=True), required=True
    )
    id = fields.UUID(required=False, dump_default=False, missing=None)

    @post_load
    def deserialize(self, data, **kwargs) -> ContentEvaluationResult:
        """
        Converts the barely typed data dictionary into an actual ContentEvaluationResult

        :param data:
        :param kwargs:
        :return:
        """
        result = ContentEvaluationResult(**data)
        for rc_key, rc_value in result.requirement_constraints.items():
            if not isinstance(rc_value, ConditionFulfilledValue):
                for enum_value in ConditionFulfilledValue:
                    if str(rc_value).upper() == enum_value.value:
                        result.requirement_constraints[rc_key] = ConditionFulfilledValue(enum_value.value)
                        break
        return result



""" Test for the evaluation of the conditions tests (Mussfeldprüfung) """
from typing import Dict, Optional

import pytest  # type:ignore[import]

from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue as cfv
from ahbicht.expressions.condition_nodes import (
    ConditionNode,
    EvaluatedComposition,
    Hint,
    RequirementConstraint,
    UnevaluatedFormatConstraint,
)
from ahbicht.expressions.requirement_constraint_expression_evaluation import evaluate_requirement_constraint_tree


class TestRequirementConstraintEvaluation:
    """Test for the evaluation of the conditions tests (Mussfeldprüfung)"""

    # some valid nodes for easier referencing
    _rc_1 = RequirementConstraint(condition_key="1", conditions_fulfilled=cfv.FULFILLED)
    _rc_2 = RequirementConstraint(condition_key="2", conditions_fulfilled=cfv.UNFULFILLED)
    _rc_3 = RequirementConstraint(condition_key="3", conditions_fulfilled=cfv.FULFILLED)
    _rc_4 = RequirementConstraint(condition_key="4", conditions_fulfilled=cfv.UNFULFILLED)
    _rc_101 = RequirementConstraint(condition_key="101", conditions_fulfilled=cfv.UNKNOWN)
    _rc_102 = RequirementConstraint(condition_key="102", conditions_fulfilled=cfv.UNKNOWN)
    _h_501 = Hint(condition_key="501", hint="[501] Hinweis: Foo")
    _h_502 = Hint(condition_key="502", hint="[502] Hinweis: Bar")
    _fc_950 = UnevaluatedFormatConstraint(condition_key="950")
    _fc_951 = UnevaluatedFormatConstraint(condition_key="951")
    _fc_987 = UnevaluatedFormatConstraint(condition_key="987")
    _fc_988 = UnevaluatedFormatConstraint(condition_key="988")

    @pytest.mark.parametrize(
        "expression, expected_resulting_conditions_fulfilled",
        [
            pytest.param("[1]", cfv.FULFILLED),
            pytest.param("[2]", cfv.UNFULFILLED),
            pytest.param("[1]U[3]", cfv.FULFILLED),
            pytest.param("[1]U[2]", cfv.UNFULFILLED),
            pytest.param("[2]U[1]", cfv.UNFULFILLED),
            pytest.param("[2]U[4]", cfv.UNFULFILLED),
            pytest.param("[1]O[3]", cfv.FULFILLED),
            pytest.param("[1]O[2]", cfv.FULFILLED),
            pytest.param("[2]O[1]", cfv.FULFILLED),
            pytest.param("[2]O[4]", cfv.UNFULFILLED),
            pytest.param("[1]X[3]", cfv.UNFULFILLED),
            pytest.param("[1]X[2]", cfv.FULFILLED),
            pytest.param("[2]X[1]", cfv.FULFILLED),
            pytest.param("[2]X[4]", cfv.UNFULFILLED),
            # Tests 'and before or'
            pytest.param("[2]U[4]O[1]", cfv.FULFILLED),
            pytest.param("[1]O[2]U[4]", cfv.FULFILLED),
            pytest.param("[2]U[1]O[3]", cfv.FULFILLED),
            pytest.param("[2]O[1]U[3]", cfv.FULFILLED),
            pytest.param("[2]O[1]U[4]", cfv.UNFULFILLED),
            pytest.param("[1]U[2]U[3]U[1]", cfv.UNFULFILLED),
            pytest.param("[1]U[3]U[2]U[1]", cfv.UNFULFILLED),
            # a very long one
            pytest.param("[1]U[2]O[1]U[1]U[2]O[2]O[1]", cfv.FULFILLED),
            # with brackets
            pytest.param("([2]U[4])O[1]", cfv.FULFILLED),
            pytest.param("[2]U([4]O[1])", cfv.UNFULFILLED),
        ],
    )
    def test_evaluate_condition_expression_with_valid_conditions_fulfilled(
        self, expression: str, expected_resulting_conditions_fulfilled: cfv
    ):
        """
        Tests that valid strings are parsed as expected.
        Odd condition_keys are True, even condition_keys are False
        """
        input_values = {"1": self._rc_1, "2": self._rc_2, "3": self._rc_3, "4": self._rc_4}

        parsed_tree = parse_condition_expression_to_tree(expression)

        result: ConditionNode = evaluate_requirement_constraint_tree(parsed_tree, input_values)
        assert result.conditions_fulfilled == expected_resulting_conditions_fulfilled

    @pytest.mark.parametrize(
        "expression, input_values, expected_error",
        [
            pytest.param("[1]", {"1": "no_boolean"}, "Please make sure that the passed values are ConditionNodes."),
            pytest.param(
                "[1]U[2]",
                {"1": "no_boolean", "2": True},
                "Please make sure that the passed values are ConditionNodes.",
            ),
            pytest.param(
                "[1]",
                {"1": EvaluatedComposition(conditions_fulfilled=cfv.FULFILLED)},
                "Please make sure that the passed values are ConditionNodes of the type RequirementConstraint, "
                "Hint or FormatConstraint.",
            ),
            # no value for [2]
            pytest.param(
                "[1]U[2]",
                {"1": RequirementConstraint(condition_key="1", conditions_fulfilled=cfv.FULFILLED)},
                "Please make sure that the input values contain all necessary condition_keys.",
            ),
        ],
    )
    def test_evaluate_condition_expression_with_invalid_values(
        self, expression: str, input_values: dict, expected_error: str
    ):
        """Tests that an error is raised when trying to pass invalid values."""
        parsed_tree = parse_condition_expression_to_tree(expression)

        with pytest.raises(ValueError) as excinfo:
            evaluate_requirement_constraint_tree(parsed_tree, input_values)

        assert expected_error in str(excinfo.value)

    @pytest.mark.parametrize(
        "expression, expected_resulting_conditions_fulfilled, expected_resulting_hint",
        [
            pytest.param("[1]", cfv.FULFILLED, None),
            pytest.param(
                "[1]U[501]", cfv.FULFILLED, "[501] Hinweis: Foo"
            ),  # e.g. [1] Segment ist genau einmal je Vorgang anzugeben,
            # dann mit [501] Hinweis: "Verwendung der ID der Messlokation"
            pytest.param("[501]U[1]", cfv.FULFILLED, "[501] Hinweis: Foo"),
            pytest.param("[2]U[501]", cfv.UNFULFILLED, None),
            pytest.param("[501]U[502]", cfv.NEUTRAL, "[501] Hinweis: Foo und [502] Hinweis: Bar"),
            pytest.param("[501]O[502]", cfv.NEUTRAL, "[501] Hinweis: Foo oder [502] Hinweis: Bar"),
            pytest.param(
                "[501]X[502]",
                cfv.NEUTRAL,
                "Entweder ([501] Hinweis: Foo) oder ([502] Hinweis: Bar)",
            ),
            pytest.param("[1]U[501]O[2]U[502]", cfv.FULFILLED, "[501] Hinweis: Foo"),
            pytest.param("[1]U[501]U[2]U[502]", cfv.UNFULFILLED, None),
            pytest.param("[1]U[501]U[3]U[502]", cfv.FULFILLED, "[501] Hinweis: Foo und [502] Hinweis: Bar"),
            # two neutral elements combined
            pytest.param("[1]U([501]O[502])", cfv.FULFILLED, "[501] Hinweis: Foo oder [502] Hinweis: Bar"),
            pytest.param(
                "[1]U([501]X[502])",
                cfv.FULFILLED,
                "Entweder ([501] Hinweis: Foo) oder ([502] Hinweis: Bar)",
            ),
            pytest.param("[1]U([501]U[502])", cfv.FULFILLED, "[501] Hinweis: Foo und [502] Hinweis: Bar"),
            pytest.param("[2]U([501]O[502])", cfv.UNFULFILLED, None),
            pytest.param("[2]U([501]X[502])", cfv.UNFULFILLED, None),
            pytest.param("[2]U([501]U[502])", cfv.UNFULFILLED, None),
        ],
    )
    def test_hints_with_valid_values(
        self,
        expression: str,
        expected_resulting_conditions_fulfilled: cfv,
        expected_resulting_hint: str,
    ):
        """Test valid expressions with Hints/Hinweise."""

        input_values = {
            "1": self._rc_1,
            "2": self._rc_2,
            "3": self._rc_3,
            "501": self._h_501,
            "502": self._h_502,
        }

        parsed_tree = parse_condition_expression_to_tree(expression)
        result: ConditionNode = evaluate_requirement_constraint_tree(parsed_tree, input_values)  # type:ignore[arg-type]

        assert result.conditions_fulfilled == expected_resulting_conditions_fulfilled
        assert getattr(result, "hint", None) == expected_resulting_hint

    @pytest.mark.parametrize(
        """expression, expected_resulting_conditions_fulfilled,
        expected_format_constraint_expression, expected_hint_text""",
        [
            pytest.param("[1][987]", cfv.FULFILLED, "[987]", None),  # true boolean + format constraint
            pytest.param("[987][1]", cfv.FULFILLED, "[987]", None),  # format constraint + true boolean
            # true boolean and unfulfilled constraint
            pytest.param("[2][987]", cfv.UNFULFILLED, None, None),  # false boolean + format constraint
            pytest.param("[987][2]", cfv.UNFULFILLED, None, None),  # format constraint + false boolean
            pytest.param("([1]O[2])[987]", cfv.FULFILLED, "[987]", None),  # true boolean + format constraint
            pytest.param("[987]([1]O[2])", cfv.FULFILLED, "[987]", None),  # format constraint + true boolean
            pytest.param("[501][987]", cfv.NEUTRAL, "[987]", "[501] Hinweis: Foo"),  # hint + format constraint
            pytest.param("[987][501]", cfv.NEUTRAL, "[987]", "[501] Hinweis: Foo"),  # format constraint + hint
            pytest.param("[987]U[988]", cfv.NEUTRAL, "[987] U [988]", None),  # format constraint U format constraint
            pytest.param("[987]O[988]", cfv.NEUTRAL, "[987] O [988]", None),  # format constraint O format constraint
            pytest.param("[987]X[988]", cfv.NEUTRAL, "[987] X [988]", None),  # format constraint X format constraint
            # two neutral elements combined
            pytest.param("[1]U([987]O[988])", cfv.FULFILLED, "([987] O [988])", None),
        ],
    )
    def test_format_constraints(
        self,
        expression: str,
        expected_resulting_conditions_fulfilled: cfv,
        expected_format_constraint_expression: Optional[str],
        expected_hint_text: Optional[str],
    ):
        """Test valid expressions with Format Constraints"""

        input_values = {
            "1": self._rc_1,
            "2": self._rc_2,
            "501": self._h_501,
            "987": self._fc_987,
            "988": self._fc_988,
        }
        parsed_tree = parse_condition_expression_to_tree(expression)
        result: EvaluatedComposition = evaluate_requirement_constraint_tree(
            parsed_tree, input_values  # type:ignore[arg-type]
        )
        assert isinstance(result, EvaluatedComposition)
        assert result.conditions_fulfilled == expected_resulting_conditions_fulfilled
        assert result.hint == expected_hint_text
        assert result.format_constraints_expression == expected_format_constraint_expression

    @pytest.mark.parametrize(
        """expression, expected_resulting_conditions_fulfilled""",
        [
            # and_composition
            pytest.param("[101]", cfv.UNKNOWN),
            pytest.param("[1]U[101]", cfv.UNKNOWN),
            pytest.param("[101]U[1]", cfv.UNKNOWN),
            pytest.param("[2]U[101]", cfv.UNFULFILLED),
            pytest.param("[101]U[2]", cfv.UNFULFILLED),
            pytest.param("[501]U[101]", cfv.UNKNOWN),
            pytest.param("[101]U[987]", cfv.UNKNOWN),
            pytest.param("[101]U[102]", cfv.UNKNOWN),
            # or_composition
            pytest.param("[1]O[101]", cfv.FULFILLED),
            pytest.param("[101]O[1]", cfv.FULFILLED),
            pytest.param("[2]O[101]", cfv.UNKNOWN),
            pytest.param("[101]O[2]", cfv.UNKNOWN),
            pytest.param("[101]O[102]", cfv.UNKNOWN),
            # xor_composition
            pytest.param("[1]X[101]", cfv.UNKNOWN),
            pytest.param("[101]X[1]", cfv.UNKNOWN),
            pytest.param("[2]X[101]", cfv.UNKNOWN),
            pytest.param("[101]X[2]", cfv.UNKNOWN),
            pytest.param("[101]X[102]", cfv.UNKNOWN),
        ],
    )
    def test_unknown_requirement_constraints(self, expression: str, expected_resulting_conditions_fulfilled: cfv):
        """Test valid expressions with unnkown requirement constraints"""

        input_values = {
            "1": self._rc_1,
            "2": self._rc_2,
            "101": self._rc_101,
            "102": self._rc_102,
            "501": self._h_501,
            "987": self._fc_987,
        }
        parsed_tree = parse_condition_expression_to_tree(expression)
        result: EvaluatedComposition = evaluate_requirement_constraint_tree(
            parsed_tree, input_values  # type:ignore[arg-type]
        )
        assert isinstance(result, ConditionNode)
        assert result.conditions_fulfilled == expected_resulting_conditions_fulfilled

    @pytest.mark.parametrize(
        "expression",
        [
            # or_composition
            pytest.param("[1]O[502]"),
            pytest.param("[502]O[1]"),
            pytest.param("[2]O[501]"),
            pytest.param("[501]O[2]"),
            pytest.param("[1]O[987]"),
            pytest.param("[987]O[1]"),
            pytest.param("[988]O[2]"),
            pytest.param("[2]O[988]"),
            ## unknown requirement constraint
            pytest.param("[501]O[101]"),
            pytest.param("[101]O[987]"),
            ## hint and format constraint
            pytest.param("[988]O[502]"),
            pytest.param("[501]O[987]"),
            ## neutral evaluated composition
            pytest.param("[1]O([501]U[502])"),
            pytest.param("[1]O([987]O[988])"),
            pytest.param("[1]O([987]X[988])"),
            # xor_composition
            pytest.param("[1]X[502]"),
            pytest.param("[502]X[1]"),
            pytest.param("[2]X[501]"),
            pytest.param("[501]X[2]"),
            pytest.param("[1]X[987]"),
            pytest.param("[987]X[1]"),
            pytest.param("[988]X[2]"),
            pytest.param("[2]X[988]"),
            ## unknown requirement constraint
            pytest.param("[501]X[101]"),
            pytest.param("[101]X[987]"),
            ## hint and format constraint
            pytest.param("[988]X[502]"),
            pytest.param("[501]X[987]"),
            ## neutral evaluated composition
            pytest.param("[1]X([501]U[502])"),
            pytest.param("[1]X([987]O[988])"),
            pytest.param("[1]X([987]X[988])"),
        ],
    )
    def test_hints_and_formats_with_invalid_or_xor_composition(self, expression: str):
        """Test invalid expressions with invalid or/xor_compositions."""

        input_values = {
            "1": self._rc_1,
            "2": self._rc_2,
            "101": self._rc_101,
            "501": self._h_501,
            "502": self._h_502,
            "987": self._fc_987,
            "988": self._fc_988,
        }

        parsed_tree = parse_condition_expression_to_tree(expression)

        with pytest.raises(NotImplementedError) as excinfo:
            evaluate_requirement_constraint_tree(parsed_tree, input_values)  # type:ignore[arg-type]

        assert """is not implemented as it has no useful result.""" in str(excinfo.value)

    @pytest.mark.parametrize(
        "input_values, expected_evaluated_result",
        [
            pytest.param(
                {
                    "1": RequirementConstraint(condition_key="1", conditions_fulfilled=cfv.UNFULFILLED),
                    "2": RequirementConstraint(condition_key="2", conditions_fulfilled=cfv.FULFILLED),
                    "3": RequirementConstraint(condition_key="3", conditions_fulfilled=cfv.UNFULFILLED),
                    "4": RequirementConstraint(condition_key="4", conditions_fulfilled=cfv.FULFILLED),
                },
                EvaluatedComposition(format_constraints_expression="[950]", conditions_fulfilled=cfv.FULFILLED),
            ),
            pytest.param(
                {
                    "1": RequirementConstraint(condition_key="1", conditions_fulfilled=cfv.FULFILLED),
                    "2": RequirementConstraint(condition_key="2", conditions_fulfilled=cfv.UNFULFILLED),
                    "3": RequirementConstraint(condition_key="3", conditions_fulfilled=cfv.FULFILLED),
                    "4": RequirementConstraint(condition_key="4", conditions_fulfilled=cfv.UNFULFILLED),
                },
                EvaluatedComposition(format_constraints_expression="[951]", conditions_fulfilled=cfv.FULFILLED),
            ),
            pytest.param(
                {
                    "1": RequirementConstraint(condition_key="1", conditions_fulfilled=cfv.FULFILLED),
                    "2": RequirementConstraint(condition_key="2", conditions_fulfilled=cfv.FULFILLED),
                    "3": RequirementConstraint(condition_key="3", conditions_fulfilled=cfv.FULFILLED),
                    "4": RequirementConstraint(condition_key="4", conditions_fulfilled=cfv.FULFILLED),
                },
                EvaluatedComposition(
                    format_constraints_expression="[950] O [951]",
                    conditions_fulfilled=cfv.FULFILLED,
                ),
            ),
            pytest.param(
                {
                    "1": RequirementConstraint(condition_key="1", conditions_fulfilled=cfv.UNFULFILLED),
                    "2": RequirementConstraint(condition_key="2", conditions_fulfilled=cfv.UNFULFILLED),
                    "3": RequirementConstraint(condition_key="3", conditions_fulfilled=cfv.UNFULFILLED),
                    "4": RequirementConstraint(condition_key="4", conditions_fulfilled=cfv.UNFULFILLED),
                },
                EvaluatedComposition(format_constraints_expression=None, conditions_fulfilled=cfv.UNFULFILLED),
            ),
        ],
    )
    def test_format_constraints_example_from_allgemeine_festlegungen(
        self,
        input_values: Dict[str, ConditionNode],
        expected_evaluated_result: EvaluatedComposition,
    ):
        """Test the example from allgemeine Festlegungen"""
        input_values["950"] = self._fc_950
        input_values["951"] = self._fc_951
        parsed_tree = parse_condition_expression_to_tree("([950] ([2] U [4])) O ([951] ([1] U [3]))")
        actual: EvaluatedComposition = evaluate_requirement_constraint_tree(
            parsed_tree, input_values  # type:ignore[arg-type]
        )
        assert isinstance(actual, EvaluatedComposition)
        assert actual == expected_evaluated_result



"""
Schemata for the JSON serialization of expressions.
"""

from typing import Optional, Union

from lark import Token, Tree
from marshmallow import Schema, fields, post_dump, post_load, pre_dump, pre_load

# in the classes/schemata we don't care about if there aren't enough public versions.
# We also don't care about unused kwargs, or no self-use.
# pylint: disable=too-few-public-methods,unused-argument,no-self-use


class _TokenOrTree:
    """
    A class that is easily serializable as dictionary and allows us to _not_ use the marshmallow-union package.
    """

    def __init__(self, token: Optional[Token] = None, tree: Optional[Tree] = None):
        self.token = token
        self.tree = tree

    token: Optional[Token]
    tree: Optional[Tree]


class _TokenOrTreeSchema(Schema):
    """
    A schema that represents data of the kind Union[str,Tree]
    There is a python package for that: https://github.com/adamboche/python-marshmallow-union
    but is only has 15 stars; not sure if it's worth the dependency
    """

    token = fields.Nested(
        lambda: TokenSchema(), allow_none=True, required=False  # pylint: disable=unnecessary-lambda
    )  # fields.String(dump_default=False, required=False, allow_none=True)
    # disable unnecessary lambda warning because of circular imports
    tree = fields.Nested(
        lambda: TreeSchema(), dump_default=False, required=False, allow_none=True  # pylint: disable=unnecessary-lambda
    )

    @post_load
    def deserialize(self, data, **kwargs) -> Union[str, Tree, Token]:
        """
        convert a dictionary back to a string, Tree or Token
        :param data:
        :param kwargs:
        :return:
        """
        if "tree" in data and data["tree"]:
            if not isinstance(data["tree"], Tree):
                return Tree(**data["tree"])
            return data["tree"]
        if "token" in data and data["token"]:
            return data["token"]
        return data

    @pre_dump
    def prepare_tree_for_serialization(self, data, **kwargs) -> _TokenOrTree:
        """
        Create a string of tree object
        :param data:
        :param kwargs:
        :return:
        """
        if isinstance(data, Tree):
            return _TokenOrTree(token=None, tree=data)
        if isinstance(data, Token):
            return _TokenOrTree(token=data, tree=None)
        raise NotImplementedError(f"Data type of {data} is not implemented for JSON serialization")


class TokenSchema(Schema):
    """
    A schema that represents a lark Token and allows (de-)serializing is as JSON
    """

    value = fields.String(dump_default=False, allow_none=True)
    type = fields.String(dump_default=False, data_key="type")

    @post_load
    def deserialize(self, data, **kwargs) -> Token:
        """
        converts the barely typed data dictionary into an actual Tree
        :param data:
        :param kwargs:
        :return:
        """
        return Token(data["type"], data["value"])


class TreeSchema(Schema):
    """
    A schema that represents a lark tree and allows (de-)serializing it as JSON
    """

    data = fields.String(data_key="type")  # for example 'or_composition', 'and_composition', 'condition_key'
    # disable lambda warning. I don't know how to resolve this circular imports
    children = fields.List(fields.Nested(lambda: _TokenOrTreeSchema()))  # pylint: disable=unnecessary-lambda

    @post_load
    def deserialize(self, data, **kwargs) -> Tree:
        """
        converts the barely typed data dictionary into an actual Tree
        :param data:
        :param kwargs:
        :return:
        """
        return Tree(**data)


def _compress(data: dict) -> dict:
    """
    a function that "throws away" unnecessary data.
    The price we pay is that we loose the ability to easily deserialize the result.
    But if we're only interested in a simple tree that's fine.
    """
    if (
        "children" in data
        and "type" in data
        and (data["type"].endswith("_composition") or data["type"].endswith("_expression"))
    ):
        return {data["type"]: [_compress(child) for child in data["children"]]}
    if "tree" in data and "token" in data and data["token"] is None:
        return _compress(data["tree"])
    if "tree" in data and "token" in data and data["tree"] is None:
        return _compress(data["token"])
    if "type" in data and "children" in data:  # and data["type"] in {"MODAL_MARK", "condition_key"}:
        return _compress(data["children"][0]["token"]["value"])
    if "type" in data and data["type"] in {"MODAL_MARK"}:
        return data["value"]
    return data


class ConciseTreeSchema(TreeSchema):
    """
    A tree schema that leads to significantly more compact output but can only be used for serialization
    (not deserialization)
    """

    @post_dump
    def serialize(self, data, **kwargs) -> dict:
        """
        Serialize as compact/concise tree
        """
        return _compress(data)

    @pre_load
    def deserialize(self, data, **kwargs) -> Tree:
        """
        This is out of scope
        """
        raise NotImplementedError("Deserializing a compact tree is not supported.")



"""
This module implements the base class for the condition check Transformers
that evaluate trees build from the condition_expression_parser.
"""

from abc import ABC, abstractmethod
from typing import Generic, Mapping, TypeVar

from lark import Token, Transformer, v_args

# All non-abstract transformers obey the following generic type constraints:
#   1. The methods of each transformer all have the same accepted argument type ("TSupportedArgumentNodeType").
#       This type is arbitrary but fixed per Transformer.
#   2. Also all of the methods of one transformer return objects of the same type ("TSupportedReturnType").
TSupportedArgumentNodeType = TypeVar("TSupportedArgumentNodeType")  # bound=ConditionNode)
# bound does not work because:
# error: Type argument "ahbicht.expressions.condition_nodes.EvaluatedFormatConstraint" of "BaseTransformer" must be a
# subtype of "ahbicht.expressions.condition_nodes.ConditionNode"  [type-var]
TSupportedReturnType = TypeVar("TSupportedReturnType")


@v_args(inline=True)  # Children are provided as *args instead of a list argument
class BaseTransformer(Transformer, ABC, Generic[TSupportedArgumentNodeType, TSupportedReturnType]):
    """
    Transformer that evaluates the trees built from the format constraint expressions.
    The input are the evaluated format constraint conditions in the form of ConditionNodes.

    After two nodes are evaluated in a composition, the resulting node has the node type EvalutatedComposition.
    The return value is a ConditionNode whose attribute `format_constraint_fulfilled` describes whether
    the format constraint expression is fulfiled or not.
    """

    def __init__(self, input_values: Mapping[str, TSupportedArgumentNodeType]):
        """
        The input are the evaluated format constraint conditions in the form of ConditionNodes.
        :param input_values: something that maps a condition key (str) onto an argument
        """
        super().__init__()
        self.input_values = input_values

    def condition_key(self, token: Token) -> TSupportedArgumentNodeType:
        """Returns ConditionNode of condition_key"""
        try:
            condition_key = self.input_values[token.value]
        except KeyError as key_err:
            raise ValueError(
                "Please make sure that the input values contain all necessary condition_keys."
            ) from key_err
        return condition_key

    @abstractmethod
    def and_composition(
        self, left: TSupportedArgumentNodeType, right: TSupportedArgumentNodeType
    ) -> TSupportedReturnType:
        """Evaluates logical and_composition"""

        raise NotImplementedError("Has to be implemented by inheriting class.")

    @abstractmethod
    def or_composition(
        self, left: TSupportedArgumentNodeType, right: TSupportedArgumentNodeType
    ) -> TSupportedReturnType:
        """Evaluates logical (inclusive) or_composition"""

        raise NotImplementedError("Has to be implemented by inheriting class.")

    @abstractmethod
    def xor_composition(
        self, left: TSupportedArgumentNodeType, right: TSupportedArgumentNodeType
    ) -> TSupportedReturnType:
        """Evaluates exclusive xor_composition"""

        raise NotImplementedError("Has to be implemented by inheriting class.")



"""
This module contains classes that are returned by mappers, meaning they contain a mapping.
"""
from typing import Optional

import attr
from marshmallow import Schema, fields, post_load
from marshmallow_enum import EnumField  # type:ignore[import]

from ahbicht.edifact import EdifactFormat


# pylint:disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class ConditionKeyConditionTextMapping:
    """
    maps a condition from a specified EDIFACT format onto a text as it is found in the AHB.
    """

    edifact_format: EdifactFormat = attr.ib(
        validator=attr.validators.instance_of(EdifactFormat)
    )  #: the format in which the condition is used; f.e. 'UTILMD'
    condition_key: str = attr.ib(
        validator=attr.validators.instance_of(str)
    )  #: the key of the condition without square brackets; f.e. '78'
    condition_text: Optional[str] = attr.ib(
        default=None
        # pylint:disable=line-too-long
    )  #: the description of the condition as in the AHB; None if unknown; f.e. 'Wenn SG4 STS+7++E02 (Transaktionsgrund: Einzug/Neuanlage)  nicht vorhanden'.


class ConditionKeyConditionTextMappingSchema(Schema):
    """
    A schema to (de-)serialize :class:`.ConditionKeyConditionTextMapping` s
    """

    edifact_format = EnumField(EdifactFormat)
    condition_key = fields.String()
    condition_text = fields.String(load_default=None)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> ConditionKeyConditionTextMapping:
        """
        Converts the barely typed data dictionary into an actual :class:`.ConditionKeyConditionTextMapping`
        """
        return ConditionKeyConditionTextMapping(**data)


# pylint:disable=too-few-public-methods
@attr.s(auto_attribs=True, kw_only=True)
class PackageKeyConditionExpressionMapping:
    """
    maps a package key from a specified EDIFACT format onto a (not yet parsed) condition expression as it is found in
    the AHB.
    """

    edifact_format: EdifactFormat = attr.ib(
        validator=attr.validators.instance_of(EdifactFormat)
    )  #: the format in which the package is used; f.e. 'UTILMD'
    package_key: str = attr.ib(
        validator=attr.validators.instance_of(str)
    )  #: the key of the package without square brackets but with trailing P; f.e. '10P'
    package_expression: Optional[str] = attr.ib(
        default=None
    )  #: the expression for which the package is a shortcut; None if unknown f.e. '[20] ∧ [244]'


class PackageKeyConditionExpressionMappingSchema(Schema):
    """
    A schema to (de-)serialize :class:`.PackageKeyConditionExpressionMapping` s
    """

    edifact_format = EnumField(EdifactFormat)
    package_key = fields.String()
    package_expression = fields.String(load_default=None)

    # pylint:disable=unused-argument,no-self-use
    @post_load
    def deserialize(self, data, **kwargs) -> PackageKeyConditionExpressionMapping:
        """
        Converts the barely typed data dictionary into an actual :class:`.PackageKeyConditionExpressionMapping`
        """
        return PackageKeyConditionExpressionMapping(**data)



"""
Tests that the code can handle RC/FC evaluators that have both async and sync methods.
"""
from unittest import mock

import inject
import pytest  # type:ignore[import]

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData
from ahbicht.content_evaluation.fc_evaluators import FcEvaluator
from ahbicht.content_evaluation.rc_evaluators import DictBasedRcEvaluator, RcEvaluator
from ahbicht.edifact import EdifactFormat, EdifactFormatVersion
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from ahbicht.expressions.hints_provider import DictBasedHintsProvider, HintsProvider

pytestmark = pytest.mark.asyncio


class MixedSyncAsyncRcEvaluator(RcEvaluator):
    """An RC evaluator that has both sync and async methods"""

    edifact_format = EdifactFormat.UTILMD
    edifact_format_version = EdifactFormatVersion.FV2104

    def _get_default_context(self):
        return None  # type:ignore[return-value]

    def evaluate_1(self, _):
        return ConditionFulfilledValue.FULFILLED

    async def evaluate_2(self, _):
        return ConditionFulfilledValue.UNFULFILLED


class MixedSyncAsyncFcEvaluator(FcEvaluator):
    """An FC Evaluator that has both sync and async methods"""

    edifact_format = EdifactFormat.UTILMD
    edifact_format_version = EdifactFormatVersion.FV2104

    def evaluate_901(self, _):
        return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)

    async def evaluate_902(self, _):
        return EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None)


class TestMixedSyncAsyncEvaluation:
    @pytest.mark.parametrize(
        "expression, expected_rc_fulfilled, expected_fc_fulfilled",
        [
            pytest.param("Muss ([1][901] X [2][902])", True, True),
            pytest.param("Muss ([1][902] X [2][901])", True, True),
            pytest.param("Muss ([2][902] X [2][901])", False, True),
        ],
    )
    async def test_mixed_async_non_async(
        self, expression: str, expected_rc_fulfilled: bool, expected_fc_fulfilled: bool
    ):
        fc_evaluator = MixedSyncAsyncFcEvaluator()
        rc_evaluator = MixedSyncAsyncRcEvaluator(EvaluatableData(edifact_seed={}))
        inject.clear_and_configure(
            lambda binder: binder.bind(RcEvaluator, rc_evaluator)  # type:ignore[arg-type]
            .bind(FcEvaluator, fc_evaluator)
            .bind(HintsProvider, DictBasedHintsProvider({}))
        )
        evaluation_input = "something has to be here but it's not important what"
        tree = parse_expression_including_unresolved_subexpressions(expression)
        evaluation_result = await evaluate_ahb_expression_tree(tree, entered_input=evaluation_input)
        assert (
            evaluation_result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            is expected_rc_fulfilled
        )
        assert (
            evaluation_result.format_constraint_evaluation_result.format_constraints_fulfilled is expected_fc_fulfilled
        )
        # When mocker.spy ing on the evaluate_... the inspection inside the rc/fc evaluator fails
        # https://stackoverflow.com/questions/18869141/using-python-mock-to-spy-on-calls-to-an-existing-object



""" Tests the Evaluator Factory"""
import uuid
from typing import Optional

import pytest  # type:ignore[import]
from _pytest.fixtures import SubRequest  # type:ignore[import]

from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult
from ahbicht.content_evaluation.evaluator_factory import create_and_inject_hardcoded_evaluators
from ahbicht.expressions.ahb_expression_evaluation import evaluate_ahb_expression_tree
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue, EvaluatedFormatConstraint

pytestmark = pytest.mark.asyncio


class TestEvaluatorFactory:
    """Tests, that evaluators are created and injected correctly"""

    @pytest.fixture
    def inject_content_evaluation_result(self, request: SubRequest):
        # indirect parametrization: https://stackoverflow.com/a/33879151/10009545
        content_evaluation_result = request.param
        create_and_inject_hardcoded_evaluators(content_evaluation_result=content_evaluation_result)

    @pytest.mark.parametrize(
        "inject_content_evaluation_result",
        [
            ContentEvaluationResult(
                hints={"501": "foo"},
                format_constraints={
                    "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                },
                requirement_constraints={
                    "2": ConditionFulfilledValue.FULFILLED,
                    "3": ConditionFulfilledValue.UNFULFILLED,
                },
                id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
            )
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "expression, expected_requirement_indicator, expected_format_constraint_result, expected_in_hints",
        [
            pytest.param("Muss ([2] O [3])[902]U[501]", "Muss", True, "foo"),
            pytest.param("Muss [2] O [3][902]U[501]", "Muss", True, None),
        ],
    )
    async def test_correct_injection(
        self,
        inject_content_evaluation_result,
        expression: str,
        expected_requirement_indicator: str,
        expected_format_constraint_result: bool,
        expected_in_hints: Optional[str],
    ):
        tree = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression=expression)
        assert tree is not None
        expression_evaluation_result = await evaluate_ahb_expression_tree(tree, entered_input="hello")
        assert (
            expression_evaluation_result.requirement_constraint_evaluation_result.requirement_constraints_fulfilled
            is True
        )
        assert expression_evaluation_result.requirement_indicator == expected_requirement_indicator
        assert (
            expression_evaluation_result.format_constraint_evaluation_result.format_constraints_fulfilled
            is expected_format_constraint_result
        )
        if expected_in_hints is not None:
            assert (
                expected_in_hints
                in expression_evaluation_result.requirement_constraint_evaluation_result.hints  # type:ignore[operator]
            )
        else:
            assert expression_evaluation_result.requirement_constraint_evaluation_result.hints is None



"""
This module parses a condition expression like "[59] U ([123] O [456])" into a tree structur using
the parsing library lark: https://lark-parser.readthedocs.io/en/latest/

The used terms are defined in the README_conditions.md.
"""
from typing import List, Union

from lark import Lark, Tree
from lark.exceptions import UnexpectedCharacters, UnexpectedEOF

from ahbicht.condition_node_distinction import ConditionNodeType, derive_condition_node_type
from ahbicht.content_evaluation.categorized_key_extract import CategorizedKeyExtract


def parse_condition_expression_to_tree(condition_expression: str) -> Tree:
    """
    Parse a given condition expression with the help of the here defined grammar to a lark tree.
    The grammar starts with condition keys, e.g. [45] and combines them with
    and _/or_compositions corresponding to U/O operators or without an operator (then_also_composition).
    It follows the boolean logic 'brackets before `then_also` before `and` before `xor` before `or`'.
    Whitespaces are ignored.

    :param condition_expression: str, e.g. '[45]U[502]O[1][906]'
    :return parsed_tree: Tree
    """

    grammar = """
    ?expression: expression "O"i expression -> or_composition
                | expression "∨" expression -> or_composition
                | expression "X"i expression -> xor_composition
                | expression "⊻" expression -> xor_composition
                | expression "U"i expression -> and_composition
                | expression "∧" expression -> and_composition
                | expression expression -> then_also_composition
                | brackets
                | condition_key
    ?brackets: "(" expression ")"
    condition_key: "[" INT "]"

    %import common.INT
    %import common.WS
    %ignore WS  // WS = whitespace
    """

    parser = Lark(grammar, start="expression")
    try:
        parsed_tree = parser.parse(condition_expression)
    except (UnexpectedEOF, UnexpectedCharacters, TypeError) as eof:
        raise SyntaxError(
            """Please make sure that:
             * all conditions have the form [INT]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """
        ) from eof

    return parsed_tree


def extract_categorized_keys_from_tree(
    tree_or_list: Union[Tree, List[str]], sanitize: bool = False
) -> CategorizedKeyExtract:
    """
    find different types of condition nodes inside the given tree or list of keys.
    The types are differentiated by their number range.
    See 'Allgemeine Festlegungen' from EDI@Energy.
    """
    result = CategorizedKeyExtract(
        format_constraint_keys=[], requirement_constraint_keys=[], hint_keys=[], package_keys=[]
    )
    condition_keys: List[str]
    if isinstance(tree_or_list, list):
        condition_keys = tree_or_list
    elif isinstance(tree_or_list, Tree):
        condition_keys = [
            x.value  # type:ignore[attr-defined]
            for x in tree_or_list.scan_values(
                lambda token: token.type == "INT"  # type:ignore[union-attr]
            )
        ]
    else:
        raise ValueError(f"{tree_or_list} is neither a list nor a {Tree.__name__}")
    for condition_key in condition_keys:
        condition_node_type = derive_condition_node_type(condition_key)
        if condition_node_type is ConditionNodeType.REQUIREMENT_CONSTRAINT:
            result.requirement_constraint_keys.append(condition_key)
        elif condition_node_type is ConditionNodeType.HINT:
            result.hint_keys.append(condition_key)
        elif condition_node_type is ConditionNodeType.FORMAT_CONSTRAINT:
            result.format_constraint_keys.append(condition_key)
        elif condition_node_type is ConditionNodeType.PACKAGE:
            result.package_keys.append(condition_key)
        else:
            raise NotImplementedError(f"The type '{condition_node_type}' is not implemented yet.")
    if sanitize:
        result.sanitize()
    return result


def extract_categorized_keys(condition_expression: str) -> CategorizedKeyExtract:
    """
    Parses the given condition expression and returns CategorizedKeyExtract as a template for content
    evaluation.
    """
    tree = parse_condition_expression_to_tree(condition_expression)
    return extract_categorized_keys_from_tree(tree, sanitize=True)



"""
This module contains the abstract class ConditionNode, which specifies the nodes of the parsed tree.
The type of the subclass is the decisive factor on how the respective node
is handled in the context of the compositions it is used in.
There are three possible input nodes (RequirementConstraint, Hint and FormatConstraint) and
the EvaluatedComposition node which results from a combination of two nodes (of all possible types).

The used terms are defined in the README_conditions.md.
"""
from abc import ABC
from enum import Enum
from typing import Optional, TypeVar

import attr

# pylint: disable=too-few-public-methods
from marshmallow import Schema, fields, post_load


class ConditionFulfilledValue(str, Enum):
    """
    Possible values to describe the state of a condition
    in the condition_fulfilled attribute of the ConditionNodes.
    """

    FULFILLED = "FULFILLED"  #: if condition is fulfilled
    UNFULFILLED = "UNFULFILLED"  #: if condition is not fulfilled
    UNKNOWN = "UNKNOWN"  #: if it cannot be checked if condition is fulfilled (e.g. "Wenn vorhanden")
    NEUTRAL = (
        "NEUTRAL"  #: a hint or unevaluated format constraint which does not have a status of being fulfilled or not
    )

    def __str__(self):
        return self.value

    def __or__(self, other):
        if other == ConditionFulfilledValue.NEUTRAL:  # todo: the next 8 lines are nearly identical with __and__
            return self
        if self == ConditionFulfilledValue.NEUTRAL:
            return other
        # if any single operand of the or composition is fulfilled, then the entire outcome is fulfilled, regardless
        # of the other operand
        if ConditionFulfilledValue.FULFILLED in (self, other):
            return ConditionFulfilledValue.FULFILLED
        # if no operand is fulfilled, then any single "unknown" leads to an unknown outcome
        if ConditionFulfilledValue.UNKNOWN in (self, other):
            return ConditionFulfilledValue.UNKNOWN
        return ConditionFulfilledValue.UNFULFILLED

    def __and__(self, other):
        if other == ConditionFulfilledValue.NEUTRAL:  # todo: the next 8 lines are nearly identical with __or__
            return self
        if self == ConditionFulfilledValue.NEUTRAL:
            return other
        # if any single operand in the and composition is unfulfilled, then the entire outcome is unfulfilled,
        # regardless of the other operand
        if ConditionFulfilledValue.UNFULFILLED in (self, other):
            return ConditionFulfilledValue.UNFULFILLED
        if ConditionFulfilledValue.UNKNOWN in (self, other):
            return ConditionFulfilledValue.UNKNOWN
        if self == ConditionFulfilledValue.FULFILLED and other == ConditionFulfilledValue.FULFILLED:
            return ConditionFulfilledValue.FULFILLED
        return ConditionFulfilledValue.UNFULFILLED

    def __xor__(self, other):
        if other == ConditionFulfilledValue.NEUTRAL:
            return self
        if self == ConditionFulfilledValue.NEUTRAL:
            return other
        if ConditionFulfilledValue.UNKNOWN in (self, other):
            return ConditionFulfilledValue.UNKNOWN
        if self == ConditionFulfilledValue.FULFILLED and other == ConditionFulfilledValue.FULFILLED:
            return ConditionFulfilledValue.UNFULFILLED
        if ConditionFulfilledValue.FULFILLED in (self, other):
            return ConditionFulfilledValue.FULFILLED
        return ConditionFulfilledValue.UNFULFILLED


@attr.s(auto_attribs=True, kw_only=True)
class ConditionNode(ABC):
    """
    This abstract class specifies the nodes of the parsed tree.
    The type of the subclass is the decisive factor on how the respective node
    is handled in the context of the compositions it is used in.
    """

    conditions_fulfilled: ConditionFulfilledValue = attr.ib(
        validator=attr.validators.instance_of(ConditionFulfilledValue)
    )


# TConditionNode is a type var that matches any class inheriting from ConditionNode (in contrast to Type[ConditionNode])
TConditionNode = TypeVar("TConditionNode", bound=ConditionNode)


@attr.s(auto_attribs=True, kw_only=True)
class ConditionKeyNodeMixin(ABC):
    """
    Nodes that have a condition key.
    """

    condition_key: str = attr.ib(validator=attr.validators.instance_of(str))


@attr.s(auto_attribs=True, kw_only=True)
class RequirementConstraint(ConditionNode, ConditionKeyNodeMixin):
    """
    Bedingung, with a requirement constraint, e.g. "falls SG2+IDE+CCI == EHZ"
    """


@attr.s(auto_attribs=True, kw_only=True)
class Hint(ConditionNode, ConditionKeyNodeMixin):
    """
    A so called 'Hinweis', just a hint, even if it is worded like a condition,
    e.g. "Hinweis: 'Es ist der alte MSB zu verwenden'"
    """

    conditions_fulfilled: ConditionFulfilledValue = ConditionFulfilledValue.NEUTRAL
    hint: str = attr.ib(validator=attr.validators.instance_of(str))  # an informatory text


@attr.s(auto_attribs=True, kw_only=True)
class FormatConstraint(ConditionNode, ConditionKeyNodeMixin):
    """
    This class is the base class of all format constraints. FormatConstraints describe that data have to obey certain
    rules, meaning those conditions with an outcome that does not change whether data are obligatory or not but
    validates existing data.
    """


@attr.s(auto_attribs=True, kw_only=True)
class UnevaluatedFormatConstraint(FormatConstraint):
    """
    This class is the base class of all unevaluated format constraints. They are used in the context of the
    Mussfeldprüfung where the constraints are collected but not evaluated yet.
    """

    conditions_fulfilled: ConditionFulfilledValue = ConditionFulfilledValue.NEUTRAL


@attr.s(auto_attribs=True)
class EvaluatedFormatConstraint:
    """
    This class is the base class of all evaluated format constraints. They are used in the context of the
    Mussfeldprüfung after the format constraints are evaluated to see if the format constraint expression is
    fulfilled or not.
    """

    format_constraint_fulfilled: bool = attr.ib(validator=attr.validators.instance_of(bool))
    error_message: Optional[str] = attr.ib(default=None)


class EvaluatedFormatConstraintSchema(Schema):
    """
    A schema to (de)serialize EvaluatedFormatConstraints.
    """

    format_constraint_fulfilled = fields.Boolean(required=True)
    error_message = fields.String(required=False, allow_none=True, dump_default=True)

    # pylint: disable=no-self-use, unused-argument
    @post_load
    def deserialize(self, data, **kwargs) -> EvaluatedFormatConstraint:
        """
        converts the barely typed data dictionary into an actual EvaluatedFormatConstraint
        :param data:
        :param kwargs:
        :return:
        """
        return EvaluatedFormatConstraint(**data)


# @attr.s(auto_attribs=True, kw_only=True)
# class EvaluatableFormatConstraint(FormatConstraint):
#     """
#     This class is the base class of all evaluatable format constraints. They are used in the context of the
#     data evaluation (_after_ the Mussfeldprüfung) and can be either True/False or None.
#     """

#     name: str = attr.ib(validator=attr.validators.instance_of(str))  # e.g. "MaLo-Prüfsummenprüfung" or "OBIS-Format"


# ideas for format constraints:
# + check sum (or more generally: computational evaluation effort) constraint for e.g. MaLo IDs
# + data bound constraints (date must not be further than 16 days in the future)
# + pattern constraint that check data against a regex pattern (e.g. obis)
# + date format constraint ("must obey 'yyyy-mm-dd'")


@attr.s(auto_attribs=True, kw_only=True)
class EvaluatedComposition(ConditionNode):
    """
    Node which is returned after a composition of two nodes is evaluated.
    """

    hint: Optional[str] = attr.ib(default=None)  # text from hints/notes
    format_constraints_expression: Optional[str] = attr.ib(default=None)  # an expression that consists of (initially
    # unevaluated) format constraints that the evaluated field needs to obey



"""
This module makes it possible to parse expressions including all their subexpressions, if present.
for example ahb_expressions which contain condition_expressions or condition_expressions which contain packages.
Parsing expressions that are nested into other expressions is refered to as "resolving".
"""

from lark import Transformer, Tree

from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree


def parse_expression_including_unresolved_subexpressions(expression: str, resolve_packages: bool = False) -> Tree:
    """
    Parses expressions and resolves its subexpressions,
    for example condition_expressions in ahb_expressions or packages in condition_expressions.
    :param expression: a syntactically valid ahb_expression or condition_expression
    :param resolve_packages: if true resolves also the packages in the condition_expressions
    """
    # TOOD: implement packages
    if resolve_packages:
        raise NotImplementedError("Resolving Packages is not implemented yet.")
    try:
        expression_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(expression)
        resolved_expression_tree = AhbExpressionResolverTransformer().transform(expression_tree)
    except SyntaxError as ahb_syntax_error:
        try:
            resolved_expression_tree = parse_condition_expression_to_tree(expression)
        except SyntaxError as condition_syntax_error:
            # pylint: disable=raise-missing-from
            raise SyntaxError(f"{ahb_syntax_error.msg} {condition_syntax_error.msg}")

    return resolved_expression_tree


# pylint: disable=no-self-use, invalid-name
class AhbExpressionResolverTransformer(Transformer):
    """
    Resolves the condition_expressions inside an ahb_expression.
    """

    def CONDITION_EXPRESSION(self, expression):
        """
        Replacing the expression_condition with its parsed tree.
        """
        condition_tree = parse_condition_expression_to_tree(expression.value)
        return condition_tree



import pytest  # type:ignore[import]
from lark import Token, Tree

from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions


class TestExpressionResolver:
    @pytest.mark.parametrize(
        "expression, expected_tree",
        [
            pytest.param(
                "Muss[3]U[4] Soll[5]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Muss"),
                                Tree(
                                    "and_composition",
                                    [
                                        Tree("condition_key", [Token("INT", "3")]),
                                        Tree("condition_key", [Token("INT", "4")]),
                                    ],
                                ),
                            ],
                        ),
                        Tree(
                            "single_requirement_indicator_expression",
                            [Token("MODAL_MARK", "Soll"), Tree("condition_key", [Token("INT", "5")])],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "X[504]O[6]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("PREFIX_OPERATOR", "X"),
                                Tree(
                                    "or_composition",
                                    [
                                        Tree("condition_key", [Token("INT", "504")]),
                                        Tree("condition_key", [Token("INT", "6")]),
                                    ],
                                ),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "[905]([504]U[6])",
                Tree(
                    "then_also_composition",
                    [
                        Tree("condition_key", [Token("INT", "905")]),
                        Tree(
                            "and_composition",
                            [
                                Tree("condition_key", [Token("INT", "504")]),
                                Tree("condition_key", [Token("INT", "6")]),
                            ],
                        ),
                    ],
                ),
            ),
        ],
    )
    def test_expression_resolver_valid(self, expression: str, expected_tree: Tree):
        actual_tree = parse_expression_including_unresolved_subexpressions(expression)
        assert actual_tree == expected_tree

    @pytest.mark.parametrize(
        "expression",
        [
            pytest.param(
                "foo",
            ),
        ],
    )
    def test_expression_resolver_failing(self, expression: str):
        with pytest.raises(SyntaxError) as excinfo:
            parse_expression_including_unresolved_subexpressions(expression)

        assert """Please make sure that the ahb_expression starts with a requirement indicator \
(i.e Muss/M, Soll/S, Kann/K, X, O, U) and the condition expressions consist of only \
the following characters: [ ] ( ) U ∧ O ∨ X ⊻ and digits.""" in str(
            excinfo.value
        )

        assert """Please make sure that:
             * all conditions have the form [INT]
             * no conditions are empty
             * all compositions are combined by operators 'U'/'O'/'X' or without an operator
             * all open brackets are closed again and vice versa
             """ in str(
            excinfo.value
        )



"""
This module manages EDIFACT related stuff. It's basically a helper module to avoid stringly typed parameters.
"""
import re
from enum import Enum
from typing import Dict, Optional

pruefidentifikator_pattern = re.compile(r"^[1-9]\d{4}$")


# pylint: disable=too-few-public-methods
class EdifactFormat(str, Enum):
    """
    existing EDIFACT formats
    """

    APERAK = "APERAK"
    COMDIS = "COMDIS"  #: communication dispute
    IFTSTA = "IFTSTA"  #: Multimodaler Statusbericht
    INSRPT = "INSRPT"  #: Prüfbericht
    INVOIC = "INVOIC"  #: invoice
    MSCONS = "MSCONS"  #: meter readings
    ORDCHG = "ORDCHG"  #: changing an order
    ORDERS = "ORDERS"  #: orders
    ORDRSP = "ORDRSP"  #: orders response
    PRICAT = "PRICAT"  #: price catalogue
    QUOTES = "QUOTES"  #: quotes
    REMADV = "REMADV"  #: zahlungsavis
    REQOTE = "REQOTE"  #: request quote
    PARTIN = "PARTIN"  #: market partner data
    UTILMD = "UTILMD"  #: utilities master data
    UTILTS = "UTILTS"  #: formula

    def __str__(self):
        return self.value


_edifact_mapping: Dict[str, EdifactFormat] = {
    "99": EdifactFormat.APERAK,
    "29": EdifactFormat.COMDIS,
    "21": EdifactFormat.IFTSTA,
    "23": EdifactFormat.INSRPT,
    "31": EdifactFormat.INVOIC,
    "13": EdifactFormat.MSCONS,
    "39": EdifactFormat.ORDCHG,
    "17": EdifactFormat.ORDERS,
    "19": EdifactFormat.ORDRSP,
    "27": EdifactFormat.PRICAT,
    "15": EdifactFormat.QUOTES,
    "33": EdifactFormat.REMADV,
    "37": EdifactFormat.PARTIN,
    "11": EdifactFormat.UTILMD,
    "25": EdifactFormat.UTILTS,
}


class EdifactFormatVersion(str, Enum):
    """
    One format version refers to the period in which an AHB is valid.
    """

    FV2104 = "FV2104"  #: valid from 2021-04-01 until 2021-10-01
    FV2110 = "FV2110"  #: valid from 2021-10-01 until 2022-04-01
    FV2204 = "FV2204"  #: valid from 2022-04-01 onwards ("MaKo 2022")

    def __str__(self):
        return self.value


def pruefidentifikator_to_format(pruefidentifikator: str) -> Optional[EdifactFormat]:
    """
    returns the format corresponding to a given pruefi
    """
    if not pruefidentifikator:
        raise ValueError("The pruefidentifikator must not be falsy")
    if not pruefidentifikator_pattern.match(pruefidentifikator):
        raise ValueError(f"The pruefidentifikator '{pruefidentifikator}' is invalid.")
    try:
        return _edifact_mapping[pruefidentifikator[:2]]
    except KeyError:
        return None



"""
Evaluators are classes that evaluate AHB conditions, meaning: Based on a condition key and a message they return either
FULFILLED (true), UNFULFILLED (false) or NEUTRAL (None). Their results are used as input for the condition validation of
the entire message.
"""
import inspect
from abc import ABC
from typing import Callable, Optional

from ahbicht.edifact import EdifactFormat, EdifactFormatVersion


# pylint: disable=no-self-use, too-few-public-methods
class Evaluator(ABC):
    """
    Base of all evaluators.
    To implement a content_evaluation create or update an inheriting class that has edifact_format and
    edifact_format_version set accordingly. Then create a method named "evaluate_123" where "123" is the condition key
    of the condition it evaluates.
    """

    # define some common attributes. They will be needed to find the correct validator for each use case.
    edifact_format: EdifactFormat = NotImplementedError(  # type:ignore[assignment]
        "The inheriting class needs to define a format to which it is applicable."
    )

    edifact_format_version: EdifactFormatVersion = NotImplementedError(  # type:ignore[assignment]
        "The inheriting class needs to define a format version."
    )

    def get_evaluation_method(self, condition_key: str) -> Optional[Callable]:
        """
        Returns the method that evaluates the condition with key condition_key
        :param condition_key: unique key of the condition, e.g. "59"
        :return: The method that can be used for content_evaluation; None if no such method is implemented.
        """
        for candidate in inspect.getmembers(self, inspect.ismethod):
            # a candidate is a tuple of a string (index 0, name of the method) and the bound method itself (index 1)
            if candidate[0] == f"evaluate_{condition_key}":
                return candidate[1]
        return None



"""
Tests that the parsed trees are JSON serializable
"""
import json
import uuid
from typing import TypeVar

import pytest  # type:ignore[import]
from lark import Token, Tree
from marshmallow import Schema, ValidationError

from ahbicht.content_evaluation.categorized_key_extract import CategorizedKeyExtract, CategorizedKeyExtractSchema
from ahbicht.content_evaluation.content_evaluation_result import ContentEvaluationResult, ContentEvaluationResultSchema
from ahbicht.edifact import EdifactFormat
from ahbicht.evaluation_results import (
    AhbExpressionEvaluationResult,
    AhbExpressionEvaluationResultSchema,
    FormatConstraintEvaluationResult,
    RequirementConstraintEvaluationResult,
)
from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions
from ahbicht.expressions.condition_expression_parser import parse_condition_expression_to_tree
from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    EvaluatedFormatConstraint,
    EvaluatedFormatConstraintSchema,
)
from ahbicht.expressions.expression_resolver import parse_expression_including_unresolved_subexpressions
from ahbicht.json_serialization.tree_schema import ConciseTreeSchema, TreeSchema
from ahbicht.mapping_results import (
    ConditionKeyConditionTextMapping,
    ConditionKeyConditionTextMappingSchema,
    PackageKeyConditionExpressionMapping,
    PackageKeyConditionExpressionMappingSchema,
)

T = TypeVar("T")


def _test_serialization_roundtrip(serializable_object: T, schema: Schema, expected_json_dict: dict) -> T:
    """
    Serializes the serializable_object using the provided schema,
    asserts, that the result is equal to the expected_json_dict
    then deserializes it again and asserts on equality with the original serializable_object
    :returns the deserialized_object
    """
    json_string = schema.dumps(serializable_object)
    assert json_string is not None
    actual_json_dict = json.loads(json_string)
    assert actual_json_dict == expected_json_dict
    deserialized_object = schema.loads(json_data=json_string)
    assert isinstance(deserialized_object, type(serializable_object))
    assert deserialized_object == serializable_object
    return deserialized_object


class TestJsonSerialization:
    @pytest.mark.parametrize(
        "tree, expected_json_dict",
        [
            pytest.param(
                Tree(
                    "or_composition",
                    [
                        Tree("condition_key", [Token("INT", "53")]),
                        Tree(
                            "and_composition",
                            [Tree("condition_key", [Token("INT", "1")]), Tree("condition_key", [Token("INT", "2")])],
                        ),
                    ],
                ),
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [{"token": {"type": "INT", "value": "53"}, "tree": None}],
                                "type": "condition_key",
                            },
                        },
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {
                                        "token": None,
                                        "tree": {
                                            "children": [{"token": {"type": "INT", "value": "1"}, "tree": None}],
                                            "type": "condition_key",
                                        },
                                    },
                                    {
                                        "token": None,
                                        "tree": {
                                            "children": [{"token": {"type": "INT", "value": "2"}, "tree": None}],
                                            "type": "condition_key",
                                        },
                                    },
                                ],
                                "type": "and_composition",
                            },
                        },
                    ],
                    "type": "or_composition",
                },
            )
        ],
    )
    def test_tree_serialization(self, tree: Tree, expected_json_dict: dict):
        _test_serialization_roundtrip(tree, TreeSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "condition_string, expected_json_dict",
        [
            pytest.param(
                "Muss [2] U ([3] O [4])[901] U [555]",
                {
                    "children": [
                        {
                            "token": None,
                            "tree": {
                                "children": [
                                    {"token": {"type": "MODAL_MARK", "value": "Muss"}, "tree": None},
                                    {
                                        "token": {
                                            "type": "CONDITION_EXPRESSION",
                                            "value": " [2] U ([3] O [4])[901] U [555]",
                                        },
                                        "tree": None,
                                    },
                                ],
                                "type": "single_requirement_indicator_expression",
                            },
                        }
                    ],
                    "type": "ahb_expression",
                },
            )
        ],
    )
    def test_single_requirement_indicator_expression_serialization(
        self, condition_string: str, expected_json_dict: dict
    ):
        tree = parse_ahb_expression_to_single_requirement_indicator_expressions(condition_string)
        _test_serialization_roundtrip(tree, TreeSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "evaluated_format_constraint, expected_json_dict",
        [
            pytest.param(
                EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                {"format_constraint_fulfilled": True, "error_message": None},
            ),
            pytest.param(
                EvaluatedFormatConstraint(format_constraint_fulfilled=False, error_message="something is wrong"),
                {"format_constraint_fulfilled": False, "error_message": "something is wrong"},
            ),
        ],
    )
    def test_evaluated_format_constraint_serialization(
        self, evaluated_format_constraint: EvaluatedFormatConstraint, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            evaluated_format_constraint, EvaluatedFormatConstraintSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "invalid_content_evaluation_result_dict",
        [
            pytest.param({}, id="empty dict"),
            pytest.param({"format_constraints": {}, "hints": {}}, id="missing requirement constraints"),
            pytest.param({"requirement_constraints": {}, "hints": {}}, id="missing format_constraints"),
        ],
    )
    def test_validation_errors_on_content_evaluation_result_deserialization(
        self, invalid_content_evaluation_result_dict: dict
    ):
        schema = ContentEvaluationResultSchema()
        with pytest.raises(ValidationError):
            schema.load(invalid_content_evaluation_result_dict)

    @pytest.mark.parametrize(
        "content_evaluation_result, expected_json_dict",
        [
            pytest.param(
                ContentEvaluationResult(
                    hints={"501": "foo", "502": "bar", "503": None},
                    format_constraints={
                        "901": EvaluatedFormatConstraint(
                            format_constraint_fulfilled=False, error_message="something is wrong"
                        ),
                        "902": EvaluatedFormatConstraint(format_constraint_fulfilled=True, error_message=None),
                    },
                    requirement_constraints={
                        "1": ConditionFulfilledValue.NEUTRAL,
                        "2": ConditionFulfilledValue.FULFILLED,
                        "3": ConditionFulfilledValue.UNFULFILLED,
                        "4": ConditionFulfilledValue.UNKNOWN,
                    },
                    id=uuid.UUID("d106f335-f663-4d14-9636-4f43a883ad26"),
                ),
                {
                    "hints": {"501": "foo", "502": "bar", "503": None},
                    "format_constraints": {
                        "902": {"format_constraint_fulfilled": True, "error_message": None},
                        "901": {"format_constraint_fulfilled": False, "error_message": "something is wrong"},
                    },
                    "requirement_constraints": {"1": "NEUTRAL", "2": "FULFILLED", "3": "UNFULFILLED", "4": "UNKNOWN"},
                    "id": "d106f335-f663-4d14-9636-4f43a883ad26",
                },
            ),
        ],
    )
    def test_content_evaluation_result_serialization(
        self, content_evaluation_result: ContentEvaluationResult, expected_json_dict: dict
    ):
        for rc_evaluation_result in content_evaluation_result.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)
        deserialized_object = _test_serialization_roundtrip(
            content_evaluation_result, ContentEvaluationResultSchema(), expected_json_dict
        )
        for rc_evaluation_result in deserialized_object.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)
        for rc_evaluation_result in content_evaluation_result.requirement_constraints.values():
            assert isinstance(rc_evaluation_result, ConditionFulfilledValue)

    @pytest.mark.parametrize(
        "ahb_expression_evaluation_result, expected_json_dict",
        [
            pytest.param(
                AhbExpressionEvaluationResult(
                    requirement_indicator="Muss",
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        error_message="hello", format_constraints_fulfilled=False
                    ),
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        hints="foo bar",
                        requirement_constraints_fulfilled=True,
                        requirement_is_conditional=True,
                        format_constraints_expression="[asd]",
                    ),
                ),
                {
                    "format_constraint_evaluation_result": {
                        "error_message": "hello",
                        "format_constraints_fulfilled": False,
                    },
                    "requirement_constraint_evaluation_result": {
                        "format_constraints_expression": "[asd]",
                        "hints": "foo bar",
                        "requirement_constraints_fulfilled": True,
                        "requirement_is_conditional": True,
                    },
                    "requirement_indicator": "Muss",
                },
            ),
            pytest.param(
                AhbExpressionEvaluationResult(
                    requirement_indicator="Muss",
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        error_message="hello", format_constraints_fulfilled=False
                    ),
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        hints="foo bar",
                        requirement_constraints_fulfilled=True,
                        requirement_is_conditional=True,
                        format_constraints_expression="[asd]",
                    ),
                ),
                {
                    "format_constraint_evaluation_result": {
                        "error_message": "hello",
                        "format_constraints_fulfilled": False,
                    },
                    "requirement_constraint_evaluation_result": {
                        "format_constraints_expression": "[asd]",
                        "hints": "foo bar",
                        "requirement_constraints_fulfilled": True,
                        "requirement_is_conditional": True,
                    },
                    "requirement_indicator": "Muss",
                },
            ),
            pytest.param(
                AhbExpressionEvaluationResult(
                    requirement_indicator="Muss",
                    format_constraint_evaluation_result=FormatConstraintEvaluationResult(
                        format_constraints_fulfilled=False
                    ),
                    requirement_constraint_evaluation_result=RequirementConstraintEvaluationResult(
                        requirement_constraints_fulfilled=True,
                        requirement_is_conditional=True,
                    ),
                ),
                {
                    "format_constraint_evaluation_result": {
                        "error_message": None,
                        "format_constraints_fulfilled": False,
                    },
                    "requirement_constraint_evaluation_result": {
                        "format_constraints_expression": None,
                        "hints": None,
                        "requirement_constraints_fulfilled": True,
                        "requirement_is_conditional": True,
                    },
                    "requirement_indicator": "Muss",
                },
                id="Minimal example",
            ),
        ],
    )
    def test_ahb_expression_evaluation_result_serialization(
        self, ahb_expression_evaluation_result: AhbExpressionEvaluationResult, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            ahb_expression_evaluation_result, AhbExpressionEvaluationResultSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "condition_key_condition_text_mapping, expected_json_dict",
        [
            pytest.param(
                ConditionKeyConditionTextMapping(
                    edifact_format=EdifactFormat.UTILMD,
                    condition_key="123",
                    condition_text="Blablabla",
                ),
                {"edifact_format": "UTILMD", "condition_key": "123", "condition_text": "Blablabla"},
            ),
        ],
    )
    def test_condition_key_condition_text_mapping_serialization(
        self, condition_key_condition_text_mapping: ConditionKeyConditionTextMapping, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            condition_key_condition_text_mapping, ConditionKeyConditionTextMappingSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "package_key_condition_expression_mapping, expected_json_dict",
        [
            pytest.param(
                PackageKeyConditionExpressionMapping(
                    edifact_format=EdifactFormat.UTILMD,
                    package_key="123P",
                    package_expression="[1] U [2] O [3]",
                ),
                {"edifact_format": "UTILMD", "package_key": "123P", "package_expression": "[1] U [2] O [3]"},
            ),
        ],
    )
    def test_package_key_condition_expression_mapping_serialization(
        self, package_key_condition_expression_mapping: PackageKeyConditionExpressionMapping, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(
            package_key_condition_expression_mapping, PackageKeyConditionExpressionMappingSchema(), expected_json_dict
        )

    @pytest.mark.parametrize(
        "categorized_key_extract, expected_json_dict",
        [
            pytest.param(
                CategorizedKeyExtract(
                    hint_keys=["501", "502", "503"],
                    format_constraint_keys=["901", "902"],
                    requirement_constraint_keys=["1", "2", "3", "4"],
                    package_keys=["17P"],
                ),
                {
                    "hint_keys": ["501", "502", "503"],
                    "format_constraint_keys": ["901", "902"],
                    "requirement_constraint_keys": ["1", "2", "3", "4"],
                    "package_keys": ["17P"],
                },
            ),
        ],
    )
    def test_categorized_key_extract_serialization(
        self, categorized_key_extract: CategorizedKeyExtract, expected_json_dict: dict
    ):
        _test_serialization_roundtrip(categorized_key_extract, CategorizedKeyExtractSchema(), expected_json_dict)

    @pytest.mark.parametrize(
        "condition_expression, expected_compact_json_dict",
        [
            pytest.param(
                "[1] U ([2] O [3])[901]",
                {"and_composition": ["1", {"then_also_composition": [{"or_composition": ["2", "3"]}, "901"]}]},
            ),
            pytest.param(
                "[3] U ([2] O [3] U [77] X [99][502])[901]",
                {
                    "and_composition": [
                        "3",
                        {
                            "then_also_composition": [
                                {
                                    "or_composition": [
                                        "2",
                                        {
                                            "xor_composition": [
                                                {"and_composition": ["3", "77"]},
                                                {"then_also_composition": ["99", "502"]},
                                            ]
                                        },
                                    ]
                                },
                                "901",
                            ]
                        },
                    ]
                },
            ),
        ],
    )
    def test_concise_tree_serialization_behaviour_for_condition_expressions(
        self, condition_expression: str, expected_compact_json_dict: dict
    ):
        tree = parse_condition_expression_to_tree(condition_expression)
        json_dict = ConciseTreeSchema().dump(tree)
        assert json_dict == expected_compact_json_dict

    @pytest.mark.parametrize(
        "ahb_expression, expected_compact_json_dict",
        [
            pytest.param(
                "Muss [1] U ([2] O [3])[901]",
                {
                    "ahb_expression": [
                        {
                            "single_requirement_indicator_expression": [
                                "Muss",
                                {
                                    "and_composition": [
                                        "1",
                                        {"then_also_composition": [{"or_composition": ["2", "3"]}, "901"]},
                                    ]
                                },
                            ]
                        }
                    ]
                },
            ),
            pytest.param(
                "Soll [3] U ([2] O [3] U [77] X [99][502])[901] Kann [43]",
                {
                    "ahb_expression": [
                        {
                            "single_requirement_indicator_expression": [
                                "Soll",
                                {
                                    "and_composition": [
                                        "3",
                                        {
                                            "then_also_composition": [
                                                {
                                                    "or_composition": [
                                                        "2",
                                                        {
                                                            "xor_composition": [
                                                                {"and_composition": ["3", "77"]},
                                                                {"then_also_composition": ["99", "502"]},
                                                            ]
                                                        },
                                                    ]
                                                },
                                                "901",
                                            ]
                                        },
                                    ]
                                },
                            ]
                        },
                        {"single_requirement_indicator_expression": ["Kann", "43"]},
                    ]
                },
            ),
        ],
    )
    def test_concise_tree_serialization_behaviour_for_ahb_expressions(
        self, ahb_expression: str, expected_compact_json_dict: dict
    ):
        tree = parse_expression_including_unresolved_subexpressions(ahb_expression)
        json_dict = ConciseTreeSchema().dump(tree)
        assert json_dict == expected_compact_json_dict



"""
Requirement Constraint (RC) Evaluators are evaluators that check if data are required under given circumstances.
Typical usecases are for example
* you must only provide a Gerätenummer if the Transaktionsgrund is f.e. 'E08'
* you must only provide an Ausbaudatum if the meter is being removed f.e. 'Z02'
"""
import asyncio
import inspect
from abc import ABC, abstractmethod
from typing import Dict, List, Optional

from ahbicht.content_evaluation.evaluationdatatypes import EvaluatableData, EvaluationContext
from ahbicht.content_evaluation.evaluators import Evaluator
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue

# pylint: disable=no-self-use


class RcEvaluator(Evaluator, ABC):
    """
    Base class of all Requirement Constraint (RC) evaluators.
    To implement a content_evaluation create or update an inheriting class that has edifact_format and
    edifact_format_version set accordingly. Then create a method named "evaluate_123" where "123" is the condition key
    of the condition it evaluates.
    """

    def __init__(self, evaluatable_data: EvaluatableData):
        """
        Instantiate any evaluator by providing some evaluatable data from a message.
        :param evaluatable_data:
        """
        if not evaluatable_data:
            raise ValueError("Evaluatable data have to be provided to any evaluator.")
        self.evaluatable_data = evaluatable_data

    @abstractmethod
    def _get_default_context(self) -> EvaluationContext:
        raise NotImplementedError("Has to be implemented in inheriting class")

    async def evaluate_single_condition(
        self, condition_key: str, context: Optional[EvaluationContext] = None
    ) -> ConditionFulfilledValue:
        """
        Evaluates the condition with the given key.
        :param context: optional content_evaluation context, if not the default context will be used
        :param condition_key: key of the condition, e.g. "78"
        :return:
        """
        evaluation_method = self.get_evaluation_method(condition_key)
        if evaluation_method is None:
            raise NotImplementedError(f"There is no content_evaluation method for condition '{condition_key}'")
        if context is None:
            context = self._get_default_context()
        if inspect.iscoroutinefunction(evaluation_method):
            return await evaluation_method(context)
        return evaluation_method(context)

    async def evaluate_conditions(
        self, condition_keys: List[str], condition_keys_with_context: Optional[Dict[str, EvaluationContext]] = None
    ) -> Dict[str, ConditionFulfilledValue]:
        """
        Validate all the conditions provided in condition_keys in their respective context.
        """
        if condition_keys_with_context is None:
            tasks = [self.evaluate_single_condition(condition_key, None) for condition_key in condition_keys]
        else:
            tasks = []
            for condition_key in condition_keys:
                if condition_key not in condition_keys_with_context:
                    tasks.append(self.evaluate_single_condition(condition_key, None))
                else:
                    tasks.append(
                        self.evaluate_single_condition(condition_key, condition_keys_with_context[condition_key])
                    )

        results = await asyncio.gather(*tasks)

        result = dict(zip(condition_keys, results))
        return result


class DictBasedRcEvaluator(RcEvaluator):
    """
    A requirement constraint evaluator that is initialized with a prefilled dictionary.
    """

    def __init__(self, results: Dict[str, ConditionFulfilledValue]):
        """
        Initialize with a dictionary that contains all the requirement constraint evaluation results.
        :param results:
        """
        super().__init__(evaluatable_data=EvaluatableData(edifact_seed=results))
        self._results: Dict[str, ConditionFulfilledValue] = results

    def _get_default_context(self) -> EvaluationContext:
        raise NotImplementedError()

    # pylint:disable=unused-argument
    async def evaluate_single_condition(
        self, condition_key: str, context: Optional[EvaluationContext] = None
    ) -> ConditionFulfilledValue:
        try:
            return self._results[condition_key]
        except KeyError as key_error:
            raise NotImplementedError(f"No result was provided for condition '{condition_key}'.") from key_error
